@@ -20,7 +20,6 @@ import java.lang.management.ManagementFactory
 import java.nio.ByteBuffer
 import java.util.concurrent.CountDownLatch
 import java.util.concurrent.locks.ReentrantLock
-import java.util.function.Consumer
 
 import javax.management.ObjectName
 import kafka.log.{AppendOrigin, Defaults, Log, LogConfig}
@@ -667,15 +666,12 @@ class TransactionStateManagerTest {
     val expiredTransactionalIds = mutable.Set.empty[String]
     appendedRecords.values.foreach { batches =>
       batches.foreach { records =>
-        records.records.forEach(new Consumer[Record] {
-          override def accept(record: Record): Unit = {
-            val transactionalId = TransactionLog.readTxnRecordKey(record.key).transactionalId
-            assertNull(record.value)
-            expiredTransactionalIds += transactionalId
-            assertEquals(Right(None), transactionManager.getTransactionState(transactionalId))
-
-          }
-        })
+        records.records.forEach { record =>
+          val transactionalId = TransactionLog.readTxnRecordKey(record.key).transactionalId
+          assertNull(record.value)
+          expiredTransactionalIds += transactionalId
+          assertEquals(Right(None), transactionManager.getTransactionState(transactionalId))
+        }
       }
     }
     expiredTransactionalIds.toSet
@@ -802,27 +798,23 @@ class TransactionStateManagerTest {
       EasyMock.capture(callbackCapture),
       EasyMock.anyObject().asInstanceOf[Option[ReentrantLock]],
       EasyMock.anyObject()
-    )).andAnswer(new IAnswer[Unit] {
-      override def answer(): Unit = {
-        callbackCapture.getValue.apply(
-          recordsCapture.getValue.map { case (topicPartition, records) =>
-            val batches = capturedAppends.getOrElse(topicPartition, {
-              val batches = mutable.Buffer.empty[MemoryRecords]
-              capturedAppends += topicPartition -> batches
-              batches
-            })
+    )).andAnswer(() => callbackCapture.getValue.apply(
+      recordsCapture.getValue.map { case (topicPartition, records) =>
+        val batches = capturedAppends.getOrElse(topicPartition, {
+          val batches = mutable.Buffer.empty[MemoryRecords]
+          capturedAppends += topicPartition -> batches
+          batches
+        })
 
-            batches += records
+        batches += records
 
-            topicPartition -> new PartitionResponse(appendError, 0L, RecordBatch.NO_TIMESTAMP, 0L)
-          }.toMap
-        )
-      }
-    }).anyTimes()
+        topicPartition -> new PartitionResponse(appendError, 0L, RecordBatch.NO_TIMESTAMP, 0L)
+      }.toMap
+    )).anyTimes()
   }
 
   private def loadTransactionsForPartitions(
-    partitionIds: Seq[Int]
+    partitionIds: Seq[Int],
   ): Unit = {
     for (partitionId <- partitionIds) {
       transactionManager.addLoadedTransactionsToCache(partitionId, 0, new Pool[String, TransactionMetadata]())
