@@ -17,22 +17,39 @@
 package kafka.security.auth
 
 import kafka.common.KafkaException
-import org.junit.{Test, Assert}
-import org.scalatest.junit.JUnitSuite
+import org.apache.kafka.common.acl.AclPermissionType
+import org.junit.Assert.assertEquals
+import org.junit.Test
+import org.scalatest.Assertions.fail
 
-class PermissionTypeTest extends JUnitSuite {
+@deprecated("Scala Authorizer API classes gave been deprecated", "Since 2.5")
+class PermissionTypeTest {
 
   @Test
   def testFromString(): Unit = {
     val permissionType = PermissionType.fromString("Allow")
-    Assert.assertEquals(Allow, permissionType)
+    assertEquals(Allow, permissionType)
 
     try {
       PermissionType.fromString("badName")
       fail("Expected exception on invalid PermissionType name.")
     } catch {
-      case e: KafkaException => // expected
+      case _: KafkaException => // expected
     }
   }
 
+  /**
+    * Test round trip conversions between org.apache.kafka.common.acl.AclPermissionType and
+    * kafka.security.auth.PermissionType.
+    */
+  @Test
+  def testJavaConversions(): Unit = {
+    AclPermissionType.values().foreach {
+      case AclPermissionType.UNKNOWN | AclPermissionType.ANY =>
+      case aclPerm =>
+        val perm = PermissionType.fromJava(aclPerm)
+        val aclPerm2 = perm.toJava
+        assertEquals(aclPerm, aclPerm2)
+    }
+  }
 }
